import { strict as assert } from 'node:assert';

import { getValue, parse, stringify } from 'markdown-code-block-meta';
import titleize from 'titleize';
import { visit } from 'unist-util-visit';

function isTab({ type, meta }, _, parent) {
  return (
    parent?.type === 'root' && type === 'code' && meta && parse(meta).has('tab')
  );
}

function matchStatement(value, target) {
  return value === target || value === target.replaceAll("'", '"');
}

function needImport(version, nodes, target) {
  return !nodes.some(
    ({ type, value, lang }) =>
      (version === 2
        ? type === 'import'
        : type === 'mdxjsEsm' ||
          (type === 'code' && lang === 'mdx-code-block')) &&
      matchStatement(value, target),
  );
}

function importStatement(version, value) {
  return version === 2
    ? { type: 'import', value }
    : { type: 'code', lang: 'mdx-code-block', value };
}

const presets = {
  js: 'JavaScript',
  ts: 'TypeScript',
  md: 'Markdown',
  mdx: 'MDX',
  json: 'JSON',
  yaml: 'YAML',
  yml: 'YAML',
  toml: 'TOML',
};

function isFirst(node, index, parent) {
  if (parent?.type !== 'root') {
    return false;
  }

  const prev = parent.children[index - 1];

  return (
    node.is === 'TabItem-begin-tag' &&
    (prev
      ? prev.is !== 'TabItem-end-tag' && prev.is !== 'Tabs-begin-tag'
      : true)
  );
}

function isLast(node, index, parent) {
  if (parent?.type !== 'root') {
    return false;
  }

  const prev = parent.children[index + 1];

  return (
    node.is === 'TabItem-end-tag' &&
    (prev
      ? prev.is !== 'TabItem-begin-tag' && prev.is !== 'Tabs-end-tag'
      : true)
  );
}

/* eslint-disable no-param-reassign */
export function autoTabs({ labels = {}, version = 2 } = {}) {
  assert(
    labels && typeof labels === 'object' && !Array.isArray(labels),
    new TypeError('`labels` should be object'),
  );
  assert([2, 3].includes(version), new TypeError('`version` should be 2 or 3'));

  const allLabels = { ...presets, ...labels };

  function jsx(object) {
    return version === 2
      ? { ...object, type: 'jsx' }
      : { ...object, type: 'code', lang: 'mdx-code-block' };
  }

  return (tree) => {
    let haveTabs = false;
    let haveTabItem = false;

    visit(tree, isTab, (node, index, parent) => {
      haveTabItem = true;

      const meta = parse(node.meta);

      const tab = getValue(meta.get('tab'));

      const { lang } = node;

      const label =
        allLabels[tab] ||
        tab ||
        allLabels[lang] ||
        titleize(lang).replace('-', ' ');

      meta.delete('tab');

      node.meta = stringify(meta);

      parent.children.splice(
        index + 1,
        0,
        jsx({
          is: 'TabItem-end-tag',
          value: '</TabItem>',
        }),
      );

      parent.children.splice(
        index,
        0,
        jsx({
          is: 'TabItem-begin-tag',
          value: `<TabItem label="${label}" value="${label}-${lang}">`,
        }),
      );
    });

    visit(tree, isFirst, (node, index, parent) => {
      haveTabs = true;
      parent.children.splice(
        index,
        0,
        jsx({
          is: 'Tabs-begin-tag',
          value: '<Tabs>',
        }),
      );
    });

    visit(tree, isLast, (node, index, parent) => {
      parent.children.splice(
        index + 1,
        0,
        jsx({
          is: 'Tabs-end-tag',
          value: '</Tabs>',
        }),
      );
    });

    if (haveTabs && needImport(version, tree.children, '@theme/Tabs')) {
      tree.children.unshift(
        importStatement(version, "import Tabs from '@theme/Tabs'"),
      );
    }

    if (haveTabItem && needImport(version, tree.children, '@theme/TabItem')) {
      tree.children.unshift(
        importStatement(version, "import TabItem from '@theme/TabItem';"),
      );
    }
  };
}
